from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    PlantWaterValueBased,
    PlantWaterValueBasedWrite,
    PlantWaterValueBasedFields,
    PlantWaterValueBasedList,
    PlantWaterValueBasedWriteList,
    PlantWaterValueBasedTextFields,
)
from cognite.powerops.client._generated.v1.data_classes._plant_water_value_based import (
    _PLANTWATERVALUEBASED_PROPERTIES_BY_FIELD,
    _create_plant_water_value_based_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .plant_water_value_based_generators import PlantWaterValueBasedGeneratorsAPI
from .plant_water_value_based_production_max_time_series import PlantWaterValueBasedProductionMaxTimeSeriesAPI
from .plant_water_value_based_production_min_time_series import PlantWaterValueBasedProductionMinTimeSeriesAPI
from .plant_water_value_based_water_value_time_series import PlantWaterValueBasedWaterValueTimeSeriesAPI
from .plant_water_value_based_feeding_fee_time_series import PlantWaterValueBasedFeedingFeeTimeSeriesAPI
from .plant_water_value_based_outlet_level_time_series import PlantWaterValueBasedOutletLevelTimeSeriesAPI
from .plant_water_value_based_inlet_level_time_series import PlantWaterValueBasedInletLevelTimeSeriesAPI
from .plant_water_value_based_head_direct_time_series import PlantWaterValueBasedHeadDirectTimeSeriesAPI
from .plant_water_value_based_query import PlantWaterValueBasedQueryAPI


class PlantWaterValueBasedAPI(NodeAPI[PlantWaterValueBased, PlantWaterValueBasedWrite, PlantWaterValueBasedList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[PlantWaterValueBased]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=PlantWaterValueBased,
            class_list=PlantWaterValueBasedList,
            class_write_list=PlantWaterValueBasedWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.generators_edge = PlantWaterValueBasedGeneratorsAPI(client)
        self.production_max_time_series = PlantWaterValueBasedProductionMaxTimeSeriesAPI(client, view_id)
        self.production_min_time_series = PlantWaterValueBasedProductionMinTimeSeriesAPI(client, view_id)
        self.water_value_time_series = PlantWaterValueBasedWaterValueTimeSeriesAPI(client, view_id)
        self.feeding_fee_time_series = PlantWaterValueBasedFeedingFeeTimeSeriesAPI(client, view_id)
        self.outlet_level_time_series = PlantWaterValueBasedOutletLevelTimeSeriesAPI(client, view_id)
        self.inlet_level_time_series = PlantWaterValueBasedInletLevelTimeSeriesAPI(client, view_id)
        self.head_direct_time_series = PlantWaterValueBasedHeadDirectTimeSeriesAPI(client, view_id)

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> PlantWaterValueBasedQueryAPI[PlantWaterValueBasedList]:
        """Query starting at plant water value baseds.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_production_max: The minimum value of the production max to filter on.
            max_production_max: The maximum value of the production max to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant water value baseds to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for plant water value baseds.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_plant_water_value_based_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_production_max,
            max_production_max,
            min_production_min,
            max_production_min,
            min_connection_losses,
            max_connection_losses,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(PlantWaterValueBasedList)
        return PlantWaterValueBasedQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        plant_water_value_based: PlantWaterValueBasedWrite | Sequence[PlantWaterValueBasedWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) plant water value baseds.

        Note: This method iterates through all nodes and timeseries linked to plant_water_value_based and creates them including the edges
        between the nodes. For example, if any of `generators` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            plant_water_value_based: Plant water value based or sequence of plant water value baseds to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new plant_water_value_based:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import PlantWaterValueBasedWrite
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_based = PlantWaterValueBasedWrite(external_id="my_plant_water_value_based", ...)
                >>> result = client.plant_water_value_based.apply(plant_water_value_based)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.plant_water_value_based.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(plant_water_value_based, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more plant water value based.

        Args:
            external_id: External id of the plant water value based to delete.
            space: The space where all the plant water value based are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete plant_water_value_based by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.plant_water_value_based.delete("my_plant_water_value_based")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.plant_water_value_based.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> PlantWaterValueBased | None: ...

    @overload
    def retrieve(
        self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> PlantWaterValueBasedList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> PlantWaterValueBased | PlantWaterValueBasedList | None:
        """Retrieve one or more plant water value baseds by id(s).

        Args:
            external_id: External id or list of external ids of the plant water value baseds.
            space: The space where all the plant water value baseds are located.

        Returns:
            The requested plant water value baseds.

        Examples:

            Retrieve plant_water_value_based by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_based = client.plant_water_value_based.retrieve("my_plant_water_value_based")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.generators_edge,
                    "generators",
                    dm.DirectRelationReference("power_ops_types", "isSubAssetOf"),
                    "outwards",
                    dm.ViewId("power_ops_core", "Generator", "1"),
                ),
            ],
        )

    def search(
        self,
        query: str,
        properties: PlantWaterValueBasedTextFields | Sequence[PlantWaterValueBasedTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PlantWaterValueBasedList:
        """Search plant water value baseds

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_production_max: The minimum value of the production max to filter on.
            max_production_max: The maximum value of the production max to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant water value baseds to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results plant water value baseds matching the query.

        Examples:

           Search for 'my_plant_water_value_based' in all text properties:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_baseds = client.plant_water_value_based.search('my_plant_water_value_based')

        """
        filter_ = _create_plant_water_value_based_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_production_max,
            max_production_max,
            min_production_min,
            max_production_min,
            min_connection_losses,
            max_connection_losses,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _PLANTWATERVALUEBASED_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PlantWaterValueBasedFields | Sequence[PlantWaterValueBasedFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: PlantWaterValueBasedTextFields | Sequence[PlantWaterValueBasedTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PlantWaterValueBasedFields | Sequence[PlantWaterValueBasedFields] | None = None,
        group_by: PlantWaterValueBasedFields | Sequence[PlantWaterValueBasedFields] = None,
        query: str | None = None,
        search_properties: PlantWaterValueBasedTextFields | Sequence[PlantWaterValueBasedTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PlantWaterValueBasedFields | Sequence[PlantWaterValueBasedFields] | None = None,
        group_by: PlantWaterValueBasedFields | Sequence[PlantWaterValueBasedFields] | None = None,
        query: str | None = None,
        search_property: PlantWaterValueBasedTextFields | Sequence[PlantWaterValueBasedTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across plant water value baseds

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_production_max: The minimum value of the production max to filter on.
            max_production_max: The maximum value of the production max to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant water value baseds to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count plant water value baseds in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.plant_water_value_based.aggregate("count", space="my_space")

        """

        filter_ = _create_plant_water_value_based_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_production_max,
            max_production_max,
            min_production_min,
            max_production_min,
            min_connection_losses,
            max_connection_losses,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _PLANTWATERVALUEBASED_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: PlantWaterValueBasedFields,
        interval: float,
        query: str | None = None,
        search_property: PlantWaterValueBasedTextFields | Sequence[PlantWaterValueBasedTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for plant water value baseds

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_production_max: The minimum value of the production max to filter on.
            max_production_max: The maximum value of the production max to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant water value baseds to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_plant_water_value_based_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_production_max,
            max_production_max,
            min_production_min,
            max_production_min,
            min_connection_losses,
            max_connection_losses,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _PLANTWATERVALUEBASED_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_production_max: float | None = None,
        max_production_max: float | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> PlantWaterValueBasedList:
        """List/filter plant water value baseds

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_production_max: The minimum value of the production max to filter on.
            max_production_max: The maximum value of the production max to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plant water value baseds to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `generators` external ids for the plant water value baseds. Defaults to True.

        Returns:
            List of requested plant water value baseds

        Examples:

            List plant water value baseds and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant_water_value_baseds = client.plant_water_value_based.list(limit=5)

        """
        filter_ = _create_plant_water_value_based_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_production_max,
            max_production_max,
            min_production_min,
            max_production_min,
            min_connection_losses,
            max_connection_losses,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.generators_edge,
                    "generators",
                    dm.DirectRelationReference("power_ops_types", "isSubAssetOf"),
                    "outwards",
                    dm.ViewId("power_ops_core", "Generator", "1"),
                ),
            ],
        )



from __future__ import annotations

import math
import re
import warnings
from itertools import groupby

from collections import Counter, defaultdict, UserList
from collections.abc import Sequence, Collection
from dataclasses import dataclass, field
from typing import Generic, Literal, Any, Iterator, Protocol, SupportsIndex, TypeVar, overload, cast, Union

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeriesList
from cognite.client.data_classes.data_modeling.instances import Instance
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.assets.data_classes._core import (
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainRelationWrite,
    PageInfo,
    GraphQLList,
    ResourcesWriteResult,
    T_DomainModel,
    T_DomainModelWrite,
    T_DomainModelWriteList,
    T_DomainModelList,
    T_DomainRelation,
    T_DomainRelationWrite,
    T_DomainRelationList,
    DomainModelCore,
    DomainRelation,
)
from cognite.powerops.client._generated.assets import data_classes


DEFAULT_LIMIT_READ = 25
DEFAULT_QUERY_LIMIT = 3
INSTANCE_QUERY_LIMIT = 1_000
# This is the actual limit of the API, we typically set it to a lower value to avoid hitting the limit.
ACTUAL_INSTANCE_QUERY_LIMIT = 10_000
IN_FILTER_LIMIT = 5_000

Aggregations = Literal["avg", "count", "max", "min", "sum"]

_METRIC_AGGREGATIONS_BY_NAME = {
    "avg": dm.aggregations.Avg,
    "count": dm.aggregations.Count,
    "max": dm.aggregations.Max,
    "min": dm.aggregations.Min,
    "sum": dm.aggregations.Sum,
}

_T_co = TypeVar("_T_co", covariant=True)


# Source from https://github.com/python/typing/issues/256#issuecomment-1442633430
# This works because str.__contains__ does not accept an object (either in typeshed or at runtime)
class SequenceNotStr(Protocol[_T_co]):
    @overload
    def __getitem__(self, index: SupportsIndex, /) -> _T_co: ...

    @overload
    def __getitem__(self, index: slice, /) -> Sequence[_T_co]: ...

    def __contains__(self, value: object, /) -> bool: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> Iterator[_T_co]: ...

    def index(self, value: Any, /, start: int = 0, stop: int = ...) -> int: ...

    def count(self, value: Any, /) -> int: ...

    def __reversed__(self) -> Iterator[_T_co]: ...


class NodeReadAPI(Generic[T_DomainModel, T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        sources: dm.ViewIdentifier | Sequence[dm.ViewIdentifier] | dm.View | Sequence[dm.View] | None,
        class_type: type[T_DomainModel],
        class_list: type[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
    ):
        self._client = client
        self._sources = sources
        self._class_type = class_type
        self._class_list = class_list
        self._view_by_read_class = view_by_read_class

    def _delete(self, external_id: str | Sequence[str], space: str) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def _retrieve(
        self,
        external_id: str,
        space: str,
        retrieve_edges: bool = False,
        edge_api_name_type_direction_view_id_penta: (
            list[tuple[EdgeAPI, str, dm.DirectRelationReference, Literal["outwards", "inwards"], dm.ViewId]] | None
        ) = None,
    ) -> T_DomainModel | None: ...

    @overload
    def _retrieve(
        self,
        external_id: SequenceNotStr[str],
        space: str,
        retrieve_edges: bool = False,
        edge_api_name_type_direction_view_id_penta: (
            list[tuple[EdgeAPI, str, dm.DirectRelationReference, Literal["outwards", "inwards"], dm.ViewId]] | None
        ) = None,
    ) -> T_DomainModelList: ...

    def _retrieve(
        self,
        external_id: str | SequenceNotStr[str],
        space: str,
        retrieve_edges: bool = False,
        edge_api_name_type_direction_view_id_penta: (
            list[tuple[EdgeAPI, str, dm.DirectRelationReference, Literal["outwards", "inwards"], dm.ViewId]] | None
        ) = None,
    ) -> T_DomainModel | T_DomainModelList | None:
        is_multiple = True
        if isinstance(external_id, str):
            node_ids = (space, external_id)
            is_multiple = False
        else:
            node_ids = [(space, ext_id) for ext_id in external_id]

        instances = self._client.data_modeling.instances.retrieve(nodes=node_ids, sources=self._sources)
        nodes = self._class_list([self._class_type.from_instance(node) for node in instances.nodes])

        if retrieve_edges and nodes:
            self._retrieve_and_set_edge_types(nodes, edge_api_name_type_direction_view_id_penta)

        if is_multiple:
            return nodes
        elif not nodes:
            return None
        else:
            return nodes[0]

    def _search(
        self,
        view_id: dm.ViewId,
        query: str,
        properties_by_field: dict[str, str],
        properties: str | Sequence[str],
        filter_: dm.Filter | None = None,
        limit: int = DEFAULT_LIMIT_READ,
    ) -> T_DomainModelList:
        if isinstance(properties, str):
            properties = [properties]

        if properties:
            properties = [properties_by_field.get(prop, prop) for prop in properties]

        nodes = self._client.data_modeling.instances.search(
            view=view_id, query=query, instance_type="node", properties=properties, filter=filter_, limit=limit
        )
        return self._class_list([self._class_type.from_instance(node) for node in nodes])

    @overload
    def _aggregate(
        self,
        view_id: dm.ViewId,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        properties_by_field: dict[str, str],
        properties: str | Sequence[str] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def _aggregate(
        self,
        view_id: dm.ViewId,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        properties_by_field: dict[str, str],
        properties: str | Sequence[str] = None,
        group_by: str | Sequence[str] | None = None,
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def _aggregate(
        self,
        view_id: dm.ViewId,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        properties_by_field: dict[str, str],
        properties: str | Sequence[str] | None = None,
        group_by: str | Sequence[str] | None = None,
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        if isinstance(group_by, str):
            group_by = [group_by]

        if group_by:
            group_by = [properties_by_field.get(prop, prop) for prop in group_by]

        if isinstance(search_properties, str):
            search_properties = [search_properties]

        if search_properties:
            search_properties = [properties_by_field.get(prop, prop) for prop in search_properties]

        if isinstance(properties, str):
            properties = [properties]

        if properties:
            properties = [properties_by_field.get(prop, prop) for prop in properties]

        if isinstance(aggregate, (str, dm.aggregations.MetricAggregation)):
            aggregate = [aggregate]

        if properties is None and (invalid := [agg for agg in aggregate if isinstance(agg, str) and agg != "count"]):
            raise ValueError(f"Cannot aggregate on {invalid} without specifying properties")

        aggregates = []
        for agg in aggregate:
            if isinstance(agg, dm.aggregations.MetricAggregation):
                aggregates.append(agg)
            elif isinstance(agg, str):
                if agg == "count" and properties is None:
                    aggregates.append(dm.aggregations.Count("externalId"))
                elif properties is None:
                    raise ValueError(f"Cannot aggregate on {agg} without specifying properties")
                else:
                    for prop in properties:
                        aggregates.append(_METRIC_AGGREGATIONS_BY_NAME[agg](prop))
            else:
                raise TypeError(f"Expected str or MetricAggregation, got {type(agg)}")

        return self._client.data_modeling.instances.aggregate(
            view=view_id,
            aggregates=aggregates,
            group_by=group_by,
            instance_type="node",
            query=query,
            properties=search_properties,
            filter=filter,
            limit=limit,
        )

    def _histogram(
        self,
        view_id: dm.ViewId,
        property: str,
        interval: float,
        properties_by_field: dict[str, str],
        query: str | None = None,
        search_properties: str | Sequence[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        property = properties_by_field.get(property, property)

        if isinstance(search_properties, str):
            search_properties = [search_properties]
        if search_properties:
            search_properties = [properties_by_field.get(prop, prop) for prop in search_properties]

        return self._client.data_modeling.instances.histogram(
            view=view_id,
            histograms=dm.aggregations.Histogram(property, interval),
            instance_type="node",
            query=query,
            properties=search_properties,
            filter=filter,
            limit=limit,
        )

    def _list(
        self,
        limit: int,
        filter: dm.Filter,
        retrieve_edges: bool = False,
        edge_api_name_type_direction_view_id_penta: (
            list[tuple[EdgeAPI, str, dm.DirectRelationReference, Literal["outwards", "inwards"]]] | None
        ) = None,
    ) -> T_DomainModelList:
        nodes = self._client.data_modeling.instances.list(
            instance_type="node", sources=self._sources, limit=limit, filter=filter
        )
        node_list = self._class_list([self._class_type.from_instance(node) for node in nodes])
        if retrieve_edges and node_list:
            self._retrieve_and_set_edge_types(node_list, edge_api_name_type_direction_view_id_penta)

        return node_list

    def _retrieve_and_set_edge_types(
        self,
        nodes: T_DomainModelList,
        edge_api_name_type_direction_view_id_penta: (
            list[tuple[EdgeAPI, str, dm.DirectRelationReference, Literal["outwards", "inwards"], dm.ViewId]] | None
        ) = None,
    ):
        for edge_type, values in groupby(edge_api_name_type_direction_view_id_penta or [], lambda x: x[2].as_tuple()):
            edges: dict[dm.EdgeId, dm.Edge] = {}
            value_list = list(values)
            for edge_api, edge_name, edge_type, direction, view_id in value_list:
                is_type = dm.filters.Equals(
                    ["edge", "type"],
                    {"space": edge_type.space, "externalId": edge_type.external_id},
                )
                if len(ids := nodes.as_node_ids()) > IN_FILTER_LIMIT:
                    filter_ = is_type
                else:
                    is_nodes = dm.filters.In(
                        ["edge", "startNode"] if direction == "outwards" else ["edge", "endNode"],
                        values=[id_.dump(camel_case=True, include_instance_type=False) for id_ in ids],
                    )
                    filter_ = dm.filters.And(is_type, is_nodes)
                result = edge_api._list(limit=-1, filter_=filter_)
                edges.update({edge.as_id(): edge for edge in result})
            edge_list = list(edges.values())
            if len(value_list) == 1:
                _, edge_name, _, direction, _ = value_list[0]
                self._set_edges(nodes, edge_list, edge_name, direction)
            else:
                # This is an 'edge' case where we have view with multiple edges of the same type.
                edge_by_end_node: dict[tuple[str, str], list[dm.Edge]] = defaultdict(list)
                for edge in edge_list:
                    node_id = edge.end_node.as_tuple() if direction == "outwards" else edge.start_node.as_tuple()
                    edge_by_end_node[node_id].append(edge)

                for no, (edge_api, edge_name, _, direction, view_id) in enumerate(value_list):
                    if not edge_by_end_node:
                        break
                    if no == len(value_list) - 1:
                        # Last edge, use all remaining nodes
                        attribute_edges = [e for e_list in edge_by_end_node.values() for e in e_list]
                    else:
                        existing = self._client.data_modeling.instances.retrieve(
                            nodes=list(edge_by_end_node), sources=view_id
                        )
                        attribute_edges = []
                        for node in existing.nodes:
                            attribute_edge = edge_by_end_node.pop(node.as_id().as_tuple(), [])
                            attribute_edges.extend(attribute_edge)

                    self._set_edges(nodes, attribute_edges, edge_name, direction)

    @staticmethod
    def _set_edges(
        nodes: Sequence[DomainModel],
        edges: Sequence[dm.Edge],
        edge_name: str,
        direction: Literal["outwards", "inwards"],
    ):
        edges_by_node: dict[tuple, list] = defaultdict(list)
        for edge in edges:
            node_id = edge.start_node.as_tuple() if direction == "outwards" else edge.end_node.as_tuple()
            edges_by_node[node_id].append(edge)

        for node in nodes:
            node_id = node.as_tuple_id()
            if node_id in edges_by_node:
                setattr(
                    node,
                    edge_name,
                    [
                        edge.end_node.external_id if direction == "outwards" else edge.start_node.external_id
                        for edge in edges_by_node[node_id]
                    ],
                )


class NodeAPI(
    Generic[T_DomainModel, T_DomainModelWrite, T_DomainModelList], NodeReadAPI[T_DomainModel, T_DomainModelList]
):
    def __init__(
        self,
        client: CogniteClient,
        sources: dm.ViewIdentifier | Sequence[dm.ViewIdentifier] | dm.View | Sequence[dm.View] | None,
        class_type: type[T_DomainModel],
        class_list: type[T_DomainModelList],
        class_write_list: type[T_DomainModelWriteList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
    ):
        super().__init__(client, sources, class_type, class_list, view_by_read_class)
        self._class_write_list = class_write_list

    def _apply(
        self, item: T_DomainModelWrite | Sequence[T_DomainModelWrite], replace: bool = False, write_none: bool = False
    ) -> ResourcesWriteResult:
        if isinstance(item, DomainModelWrite):
            instances = item.to_instances_write(self._view_by_read_class, write_none)
        else:
            instances = self._class_write_list(item).to_instances_write(self._view_by_read_class, write_none)
        result = self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )
        time_series = []
        if instances.time_series:
            time_series = self._client.time_series.upsert(instances.time_series, mode="patch")

        return ResourcesWriteResult(result.nodes, result.edges, TimeSeriesList(time_series))


class EdgeAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def _list(
        self,
        limit: int = DEFAULT_LIMIT_READ,
        filter_: dm.Filter | None = None,
    ) -> dm.EdgeList:
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=filter_)


class EdgePropertyAPI(EdgeAPI, Generic[T_DomainRelation, T_DomainRelationWrite, T_DomainRelationList]):
    def __init__(
        self,
        client: CogniteClient,
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
        class_type: type[T_DomainRelation],
        class_write_type: type[T_DomainRelationWrite],
        class_list: type[T_DomainRelationList],
    ):
        super().__init__(client)
        self._view_by_read_class = view_by_read_class
        self._view_id = view_by_read_class[class_type]
        self._class_type = class_type
        self._class_write_type = class_write_type
        self._class_list = class_list

    def _list(
        self,
        limit: int = DEFAULT_LIMIT_READ,
        filter_: dm.Filter | None = None,
    ) -> T_DomainRelationList:
        edges = self._client.data_modeling.instances.list("edge", limit=limit, filter=filter_, sources=[self._view_id])
        return self._class_list([self._class_type.from_instance(edge) for edge in edges])


@dataclass
class QueryStep:
    # Setup Variables
    name: str
    expression: dm.query.ResultSetExpression
    max_retrieve_limit: int
    select: dm.query.Select
    result_cls: type[DomainModelCore] | None = None
    is_single_direct_relation: bool = False

    # Query Variables
    cursor: str | None = None
    total_retrieved: int = 0
    results: list[Instance] = field(default_factory=list)
    last_batch_count: int = 0

    def update_expression_limit(self) -> None:
        if self.is_unlimited:
            self.expression.limit = ACTUAL_INSTANCE_QUERY_LIMIT
        else:
            self.expression.limit = max(min(INSTANCE_QUERY_LIMIT, self.max_retrieve_limit - self.total_retrieved), 0)

    @property
    def is_unlimited(self) -> bool:
        return self.max_retrieve_limit in {None, -1, math.inf}

    @property
    def is_finished(self) -> bool:
        return (
            (not self.is_unlimited and self.total_retrieved >= self.max_retrieve_limit)
            or self.cursor is None
            or self.last_batch_count == 0
            # Single direct relations are dependent on the parent node,
            # so we assume that the parent node is the limiting factor.
            or self.is_single_direct_relation
        )


class QueryBuilder(UserList, Generic[T_DomainModelList]):
    # The unique string is in case the data model has a field that ends with _\d+. This will make sure we don't
    # clean the name of the field.
    _unique_str = "a418"
    _name_pattern = re.compile(r"_a418\d+$")

    def __init__(self, result_cls: type[T_DomainModelList], nodes: Collection[QueryStep] = None):
        super().__init__(nodes or [])
        self._result_cls = result_cls

    # The dunder implementations are to get proper type hints
    def __iter__(self) -> Iterator[QueryStep]:
        return super().__iter__()

    @overload
    def __getitem__(self, item: int) -> QueryStep: ...

    @overload
    def __getitem__(self: type[QueryBuilder[T_DomainModelList]], item: slice) -> QueryBuilder[T_DomainModelList]: ...

    def __getitem__(self, item: int | slice) -> QueryStep | QueryBuilder[T_DomainModelList]:
        if isinstance(item, slice):
            return self.__class__(self.data[item])
        elif isinstance(item, int):
            return self.data[item]
        else:
            raise TypeError(f"Expected int or slice, got {type(item)}")

    def next_name(self, name: str) -> str:
        counter = Counter(self._clean_name(step.name) for step in self)
        if name in counter:
            return f"{name}_{self._unique_str}{counter[name]}"
        return name

    def _clean_name(self, name: str) -> str:
        return self._name_pattern.sub("", name)

    def reset(self):
        for expression in self:
            expression.total_retrieved = 0
            expression.cursor = None
            expression.results = []

    def update_expression_limits(self) -> None:
        for expression in self:
            expression.update_expression_limit()

    def build(self) -> dm.query.Query:
        with_ = {expression.name: expression.expression for expression in self}
        select = {expression.name: expression.select for expression in self if expression.select}
        cursors = self.cursors

        return dm.query.Query(with_=with_, select=select, cursors=cursors)

    @property
    def cursors(self) -> dict[str, str | None]:
        return {expression.name: expression.cursor for expression in self}

    def update(self, batch: dm.query.QueryResult):
        for expression in self:
            if expression.name not in batch:
                continue
            expression.last_batch_count = len(batch[expression.name])
            expression.total_retrieved += expression.last_batch_count
            expression.cursor = batch.cursors.get(expression.name)
            expression.results.extend(batch[expression.name].data)

    @property
    def is_finished(self):
        return all(expression.is_finished for expression in self)

    def unpack(self) -> T_DomainModelList:
        nodes_by_type: dict[str | None, dict[tuple[str, str], DomainModel]] = defaultdict(dict)
        edges_by_type_by_source_node: dict[tuple[str, str, str], dict[tuple[str, str], list[dm.Edge]]] = defaultdict(
            lambda: defaultdict(list)
        )
        relation_by_type_by_start_node: dict[tuple[str, str], dict[tuple[str, str], list[DomainRelation]]] = (
            defaultdict(lambda: defaultdict(list))
        )
        node_attribute_to_node_type: dict[str, str] = {}

        for step in self:
            name = step.name
            from_ = step.expression.from_

            if isinstance(step.expression, dm.query.NodeResultSetExpression) and from_:
                node_attribute_to_node_type[from_] = name

            if step.result_cls is None:  # This is a data model edge.
                for edge in step.results:
                    edge = cast(dm.Edge, edge)
                    edge_source = edge.start_node if step.expression.direction == "outwards" else edge.end_node
                    edges_by_type_by_source_node[(from_, name, step.expression.direction)][
                        (edge_source.space, edge_source.external_id)
                    ].append(edge)
            elif issubclass(step.result_cls, DomainModel):
                for node in step.results:
                    domain = step.result_cls.from_instance(node)
                    if (id_ := domain.as_tuple_id()) not in nodes_by_type[name]:
                        nodes_by_type[name][id_] = domain
            elif issubclass(step.result_cls, DomainRelation):
                for edge in step.results:
                    domain = step.result_cls.from_instance(edge)
                    relation_by_type_by_start_node[(from_, name)][domain.start_node.as_tuple()].append(domain)

            # Link direct relations
            is_direct_relation = (
                isinstance(step.expression, dm.query.NodeResultSetExpression) and from_ and from_ in nodes_by_type
            )
            if is_direct_relation:
                end_nodes = nodes_by_type[name]
                attribute_name = node_attribute_to_node_type[from_]
                for parent_node in nodes_by_type[from_].values():
                    attribute_value = getattr(parent_node, attribute_name)
                    if isinstance(attribute_value, str):
                        end_id = (parent_node.space, attribute_value)
                    elif isinstance(attribute_value, dm.NodeId):
                        end_id = attribute_value.space, attribute_value.external_id
                    else:
                        continue
                    if end_id in end_nodes:
                        setattr(parent_node, attribute_name, end_nodes[end_id])
                    else:
                        warnings.warn(f"Unpacking of query result: Could not find node with id {end_id}", stacklevel=2)

        for (node_name, node_attribute), relations_by_start_node in relation_by_type_by_start_node.items():
            for node in nodes_by_type[node_name].values():
                setattr(node, node_attribute, relations_by_start_node.get(node.as_tuple_id(), []))
            for relations in relations_by_start_node.values():
                for relation in relations:
                    edge_name = relation.edge_type.external_id.split(".")[-1]
                    if (nodes := nodes_by_type.get(edge_name)) and (
                        node := nodes.get((relation.end_node.space, relation.end_node.external_id))
                    ):
                        # Relations always have an end node.
                        relation.end_node = node

        for (node_name, node_attribute, direction), edges_by_source_node in edges_by_type_by_source_node.items():
            for node in nodes_by_type[node_name].values():
                edges = edges_by_source_node.get(node.as_tuple_id(), [])
                nodes = nodes_by_type.get(node_attribute_to_node_type.get(node_attribute), {})
                if direction == "outwards":
                    setattr(
                        node, node_attribute, [node for edge in edges if (node := nodes.get(edge.end_node.as_tuple()))]
                    )
                else:  # inwards
                    setattr(
                        node,
                        node_attribute,
                        [node for edge in edges if (node := nodes.get(edge.start_node.as_tuple()))],
                    )

        return self._result_cls(nodes_by_type[self[0].name].values())


class QueryAPI(Generic[T_DomainModelList]):
    def __init__(
        self,
        client: CogniteClient,
        builder: QueryBuilder[T_DomainModelList],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId],
    ):
        self._client = client
        self._builder = builder
        self._view_by_read_class = view_by_read_class

    def _query(self) -> T_DomainModelList:
        self._builder.reset()
        query = self._builder.build()

        while True:
            self._builder.update_expression_limits()
            query.cursors = self._builder.cursors
            batch = self._client.data_modeling.instances.query(query)
            self._builder.update(batch)
            if self._builder.is_finished:
                break
        return self._builder.unpack()


def _create_edge_filter(
    edge_type: dm.DirectRelationReference,
    start_node: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
    start_node_space: str = "IntegrationTestsImmutable",
    end_node: str | list[str] | dm.NodeId | list[dm.NodeId] | None = None,
    space_end_node: str = "IntegrationTestsImmutable",
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter:
    filters: list[dm.Filter] = [
        dm.filters.Equals(
            ["edge", "type"],
            {"space": edge_type.space, "externalId": edge_type.external_id},
        )
    ]
    if start_node and isinstance(start_node, str):
        filters.append(
            dm.filters.Equals(["edge", "startNode"], value={"space": start_node_space, "externalId": start_node})
        )
    elif start_node and isinstance(start_node, dm.NodeId):
        filters.append(
            dm.filters.Equals(
                ["edge", "startNode"], value=start_node.dump(camel_case=True, include_instance_type=False)
            )
        )
    if start_node and isinstance(start_node, list):
        filters.append(
            dm.filters.In(
                ["edge", "startNode"],
                values=[
                    (
                        {"space": start_node_space, "externalId": ext_id}
                        if isinstance(ext_id, str)
                        else ext_id.dump(camel_case=True, include_instance_type=False)
                    )
                    for ext_id in start_node
                ],
            )
        )
    if end_node and isinstance(end_node, str):
        filters.append(dm.filters.Equals(["edge", "endNode"], value={"space": space_end_node, "externalId": end_node}))
    elif end_node and isinstance(end_node, dm.NodeId):
        filters.append(
            dm.filters.Equals(["edge", "endNode"], value=end_node.dump(camel_case=True, include_instance_type=False))
        )
    if end_node and isinstance(end_node, list):
        filters.append(
            dm.filters.In(
                ["edge", "endNode"],
                values=[
                    (
                        {"space": space_end_node, "externalId": ext_id}
                        if isinstance(ext_id, str)
                        else ext_id.dump(camel_case=True, include_instance_type=False)
                    )
                    for ext_id in end_node
                ],
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["edge", "externalId"], value=external_id_prefix))
    if space and isinstance(space, str):
        filters.append(dm.filters.Equals(["edge", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["edge", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters)


class GraphQLQueryResponse:
    def __init__(self, data_model_id: dm.DataModelId):
        self._output = GraphQLList([])
        self._data_class_by_type = _GRAPHQL_DATA_CLASS_BY_DATA_MODEL_BY_TYPE[data_model_id]

    def parse(self, response: dict[str, Any]) -> GraphQLList:
        if "errors" in response:
            raise RuntimeError(response["errors"])
        _, data = list(response.items())[0]
        self._parse_item(data)
        if "pageInfo" in data:
            self._output.page_info = PageInfo.load(data["pageInfo"])
        return self._output

    def _parse_item(self, data: dict[str, Any]) -> None:
        if "items" in data:
            for item in data["items"]:
                self._parse_item(item)
        elif "__typename" in data:
            try:
                item = self._data_class_by_type[data["__typename"]].model_validate(data)
            except KeyError:
                raise ValueError(f"Could not find class for type {data['__typename']}")
            else:
                self._output.append(item)
        else:
            raise RuntimeError("Missing '__typename' in GraphQL response. Cannot determine the type of the response.")


_GRAPHQL_DATA_CLASS_BY_DATA_MODEL_BY_TYPE = {
    dm.DataModelId("power-ops-assets", "PowerAsset", "1"): {
        "PriceArea": data_classes.PriceAreaGraphQL,
        "Watercourse": data_classes.WatercourseGraphQL,
        "Plant": data_classes.PlantGraphQL,
        "Generator": data_classes.GeneratorGraphQL,
        "Reservoir": data_classes.ReservoirGraphQL,
        "TurbineEfficiencyCurve": data_classes.TurbineEfficiencyCurveGraphQL,
        "GeneratorEfficiencyCurve": data_classes.GeneratorEfficiencyCurveGraphQL,
        "BidMethod": data_classes.BidMethodGraphQL,
    },
}



from __future__ import annotations

import re

from cognite.client.data_classes import TimeSeries

from cognite.powerops.client._generated.assets import data_classes as assets
from cognite.powerops.resync import config
from cognite.powerops.resync.models._shared_v1_v2.production_model import (
    _get_single_value,
    _plant_to_inlet_reservoir_with_losses,
    head_loss_factor_fallback,
    p_max_fallback,
    p_min_fallback,
)
from cognite.powerops.resync.models.v2.production_dm import PowerAssetModelDM


def to_asset_data_model(configuration: config.ProductionConfig) -> PowerAssetModelDM:
    model = PowerAssetModelDM()
    plant_time_series_mappings_by_name = {
        mapping.plant_name: mapping for mapping in (configuration.plant_time_series_mappings or [])
    }
    start_stop_cost_time_series_by_generator = {
        mapping.generator_name: mapping.start_stop_cost
        for mapping in (configuration.generator_time_series_mappings or [])
    }
    is_generator_available = {
        mapping.generator_name: mapping.is_available for mapping in (configuration.generator_time_series_mappings or [])
    }

    for watercourse_config in configuration.watercourses:
        watercourse = assets.WatercourseApply(
            external_id=f"watercourse_{watercourse_config.name}",
            name=watercourse_config.name,
            penalty_limit=watercourse_config.shop_penalty_limit,
            plants=[],
            production_obligation=[
                TimeSeries(external_id=ext_id) for ext_id in watercourse_config.production_obligation_ts_ext_ids or []
            ],
        )
        model.watercourses.append(watercourse)

        shop_case = watercourse_config.shop_model_template

        reservoirs = []
        for reservoir_name in shop_case["model"]["reservoir"]:
            reservoir = assets.ReservoirApply(
                external_id=f"reservoir_{reservoir_name}",
                name=reservoir_name,
                **dict(
                    zip(
                        ["display_name", "ordering"],
                        watercourse_config.reservoir_display_names_and_order.get(
                            reservoir_name, (re.sub(r"\([0-9]+\)", "", reservoir_name), "999")
                        ),
                    )
                ),
            )
            reservoirs.append(reservoir)
        model.reservoirs.extend(reservoirs)

        for generator_name, generator_attributes in shop_case["model"]["generator"].items():
            start_stop_cost = start_stop_cost_time_series_by_generator.get(generator_name)
            generator_curve = generator_attributes["gen_eff_curve"]
            turbine_curves = generator_attributes["turb_eff_curves"]
            generator = assets.GeneratorApply(
                external_id=f"generator_{generator_name}",
                name=generator_name,
                penstock=int(generator_attributes.get("penstock", 1)),
                p_min=float(generator_attributes.get("p_min", 0.0)),
                start_cost=float(_get_single_value(generator_attributes.get("startcost", 0.0))),
                start_stop_cost=start_stop_cost,
                is_available_time_series=is_generator_available.get(generator_name),
                efficiency_curve=assets.GeneratorEfficiencyCurveApply(
                    external_id=f"generator_efficiency_{generator_name}",
                    power=generator_curve["x"],
                    efficiency=generator_curve["y"],
                    ref=generator_curve.get("ref"),
                ),
                turbine_curves=[
                    assets.TurbineEfficiencyCurveApply(
                        external_id=f"turbine_efficiency_{generator_name}_head_{int(turbine_curve['ref'])}",
                        head=turbine_curve["ref"],
                        flow=turbine_curve["x"],
                        efficiency=turbine_curve["y"],
                    )
                    for turbine_curve in turbine_curves
                ],
            )
            model.turbine_curves.extend(generator.turbine_curves)
            model.generators.append(generator)

        generators_by_name = {generator.name: generator for generator in model.generators}
        plants = []
        for plant_name, attributes in shop_case["model"]["plant"].items():
            display_name, order = watercourse_config.plant_display_names_and_order.get(
                plant_name, (re.sub(r"\([0-9]+\)", "", plant_name), "999")
            )
            mapping = plant_time_series_mappings_by_name.get(plant_name)
            if mapping:
                mappings = dict(
                    water_value_time_series=mapping.water_value,
                    inlet_level_time_series=mapping.inlet_reservoir_level,
                    outlet_level_time_series=mapping.outlet_reservoir_level,
                    feeding_fee_time_series=mapping.feeding_fee,
                    p_min_time_series=mapping.p_min,
                    p_max_time_series=mapping.p_max,
                    head_direct_time_series=mapping.head_direct,
                )
            else:
                mappings = {}

            all_connections = shop_case["connections"]
            all_junctions = shop_case["model"].get("junction", {})
            all_tunnels = shop_case["model"].get("tunnel", {})
            inlet_reservoir_name, connection_losses = _plant_to_inlet_reservoir_with_losses(
                plant_name, all_connections, all_junctions, all_tunnels, {r.name for r in model.reservoirs}
            )

            # TODO: In next iteration of production data model,
            # we will have to add field connection losses and regenerate pygen sdk
            plant = assets.PlantApply(
                external_id=f"plant_{plant_name}",
                name=plant_name,
                display_name=display_name,
                ordering=order,
                watercourse=watercourse,
                outlet_level=float(attributes.get("outlet_line", 0)),
                p_min=float(attributes.get("p_min", p_min_fallback)),
                p_max=float(attributes.get("p_max", p_max_fallback)),
                head_loss_factor=float(attributes.get("main_loss", [head_loss_factor_fallback])[0]),
                connection_losses=connection_losses,
                penstock_head_loss_factors={
                    str(penstock_number): float(loss_factor)
                    for penstock_number, loss_factor in enumerate(
                        attributes.get("penstock_loss", [head_loss_factor_fallback]), start=1
                    )
                },
                **mappings,
            )

            selected_reservoir = next((r for r in model.reservoirs if r.name == inlet_reservoir_name), None)
            if selected_reservoir is not None:
                plant.inlet_reservoir = selected_reservoir
            else:
                # Todo Raise Exception?
                ...
            parsed_connections = [config.Connection(**connection) for connection in all_connections]
            # Add the generators to the plant
            plant.generators = [
                g
                for connection in parsed_connections
                if (g := connection.to_from_any(generators_by_name, plant.name, "plant"))
            ] + [
                g
                for connection in parsed_connections
                if (g := connection.from_to_any(plant.name, "plant", generators_by_name))
            ]
            plants.append(plant)

            prod_area = str(next(iter(attributes["prod_area"].values())))
            price_area_name = watercourse_config.market_to_price_area[prod_area]
            price_area = assets.PriceAreaApply(
                name=price_area_name, external_id=f"price_area_{price_area_name}", timezone="Europe/Oslo"
            )
            if price_area_name not in {a.name for a in model.price_areas}:
                if price_area_name in configuration.dayahead_price_timeseries:
                    price_area.day_ahead_price = configuration.dayahead_price_timeseries[price_area_name]
                model.price_areas.append(price_area)
            price_area = next(a for a in model.price_areas if a.name == price_area_name)

            if watercourse.plants is None:
                watercourse.plants = [plant]
            else:
                watercourse.plants.append(plant)
            if price_area.plants is None:
                price_area.plants = [plant]
            else:
                price_area.plants.append(plant)
            if watercourse.name not in {w.name for w in price_area.watercourses or []}:
                if price_area.watercourses is None:
                    price_area.watercourses = [watercourse]
                else:
                    price_area.watercourses.append(watercourse)
        model.plants.extend(plants)

    return model



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.cogshop1.data_classes import (
    ModelTemplate,
    ModelTemplateApply,
    ModelTemplateApplyList,
    ModelTemplateFields,
    ModelTemplateList,
    ModelTemplateTextFields,
)
from cognite.powerops.client._generated.cogshop1.data_classes._model_template import _MODELTEMPLATE_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, IN_FILTER_LIMIT, Aggregations, TypeAPI


class ModelTemplateBaseMappingsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str], space="cogShop") -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "ModelTemplate.baseMappings"},
        )
        if isinstance(external_id, str):
            is_model_template = f.Equals(
                ["edge", "startNode"],
                {"space": space, "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_model_template)
            )

        else:
            is_model_templates = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_model_templates)
            )

    def list(
        self, model_template_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ, space="cogShop"
    ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": space, "externalId": "ModelTemplate.baseMappings"},
        )
        filters.append(is_edge_type)
        if model_template_id:
            model_template_ids = [model_template_id] if isinstance(model_template_id, str) else model_template_id
            is_model_templates = f.In(
                ["edge", "startNode"],
                [{"space": space, "externalId": ext_id} for ext_id in model_template_ids],
            )
            filters.append(is_model_templates)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class ModelTemplateAPI(TypeAPI[ModelTemplate, ModelTemplateApply, ModelTemplateList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=ModelTemplate,
            class_apply_type=ModelTemplateApply,
            class_list=ModelTemplateList,
        )
        self._view_id = view_id
        self.base_mappings = ModelTemplateBaseMappingsAPI(client)

    def apply(
        self, model_template: ModelTemplateApply | Sequence[ModelTemplateApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(model_template, ModelTemplateApply):
            instances = model_template.to_instances_apply()
        else:
            instances = ModelTemplateApplyList(model_template).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="cogShop") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> ModelTemplate: ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ModelTemplateList: ...

    def retrieve(self, external_id: str | Sequence[str]) -> ModelTemplate | ModelTemplateList:
        if isinstance(external_id, str):
            model_template = self._retrieve((self._sources.space, external_id))

            base_mapping_edges = self.base_mappings.retrieve(external_id)
            model_template.base_mappings = [edge.end_node.external_id for edge in base_mapping_edges]

            return model_template
        else:
            model_templates = self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

            base_mapping_edges = self.base_mappings.retrieve(external_id)
            self._set_base_mappings(model_templates, base_mapping_edges)

            return model_templates

    def search(
        self,
        query: str,
        properties: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ModelTemplateList:
        filter_ = _create_filter(
            self._view_id,
            version,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            watercourse_prefix,
            model,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _MODELTEMPLATE_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        group_by: ModelTemplateFields | Sequence[ModelTemplateFields] = None,
        query: str | None = None,
        search_properties: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        group_by: ModelTemplateFields | Sequence[ModelTemplateFields] | None = None,
        query: str | None = None,
        search_property: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            version,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            watercourse_prefix,
            model,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _MODELTEMPLATE_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: ModelTemplateFields,
        interval: float,
        query: str | None = None,
        search_property: ModelTemplateTextFields | Sequence[ModelTemplateTextFields] | None = None,
        version: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            version,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            watercourse_prefix,
            model,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _MODELTEMPLATE_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        version: str | list[str] | None = None,
        version_prefix: str | None = None,
        shop_version: str | list[str] | None = None,
        shop_version_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        source: str | list[str] | None = None,
        source_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> ModelTemplateList:
        filter_ = _create_filter(
            self._view_id,
            version,
            version_prefix,
            shop_version,
            shop_version_prefix,
            watercourse,
            watercourse_prefix,
            model,
            source,
            source_prefix,
            external_id_prefix,
            filter,
        )

        model_templates = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            if len(external_ids := model_templates.as_external_ids()) > IN_FILTER_LIMIT:
                base_mapping_edges = self.base_mappings.list(limit=-1)
            else:
                base_mapping_edges = self.base_mappings.list(external_ids, limit=-1)
            self._set_base_mappings(model_templates, base_mapping_edges)

        return model_templates

    @staticmethod
    def _set_base_mappings(model_templates: Sequence[ModelTemplate], base_mapping_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in base_mapping_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for model_template in model_templates:
            node_id = model_template.id_tuple()
            if node_id in edges_by_start_node:
                model_template.base_mappings = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    version: str | list[str] | None = None,
    version_prefix: str | None = None,
    shop_version: str | list[str] | None = None,
    shop_version_prefix: str | None = None,
    watercourse: str | list[str] | None = None,
    watercourse_prefix: str | None = None,
    model: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    source: str | list[str] | None = None,
    source_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if version and isinstance(version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("version"), value=version))
    if version and isinstance(version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("version"), values=version))
    if version_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("version"), value=version_prefix))
    if shop_version and isinstance(shop_version, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopVersion"), value=shop_version))
    if shop_version and isinstance(shop_version, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopVersion"), values=shop_version))
    if shop_version_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopVersion"), value=shop_version_prefix))
    if watercourse and isinstance(watercourse, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("watercourse"), value=watercourse))
    if watercourse and isinstance(watercourse, list):
        filters.append(dm.filters.In(view_id.as_property_ref("watercourse"), values=watercourse))
    if watercourse_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("watercourse"), value=watercourse_prefix))
    if model and isinstance(model, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("model"), value={"space": "cogShop", "externalId": model})
        )
    if model and isinstance(model, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("model"), value={"space": model[0], "externalId": model[1]})
        )
    if model and isinstance(model, list) and isinstance(model[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("model"), values=[{"space": "cogShop", "externalId": item} for item in model]
            )
        )
    if model and isinstance(model, list) and isinstance(model[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("model"), values=[{"space": item[0], "externalId": item[1]} for item in model]
            )
        )
    if source and isinstance(source, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("source"), value=source))
    if source and isinstance(source, list):
        filters.append(dm.filters.In(view_id.as_property_ref("source"), values=source))
    if source_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("source"), value=source_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Sequence

from cognite.client import ClientConfig, CogniteClient, data_modeling as dm
from cognite.client.data_classes import TimeSeriesList
from cognite.client.credentials import OAuthClientCredentials

from ._api.alert import AlertAPI
from ._api.bid_configuration_day_ahead import BidConfigurationDayAheadAPI
from ._api.bid_document import BidDocumentAPI
from ._api.bid_document_afrr import BidDocumentAFRRAPI
from ._api.bid_document_day_ahead import BidDocumentDayAheadAPI
from ._api.bid_matrix import BidMatrixAPI
from ._api.bid_matrix_information import BidMatrixInformationAPI
from ._api.bid_row import BidRowAPI
from ._api.date_specification import DateSpecificationAPI
from ._api.function_input import FunctionInputAPI
from ._api.function_output import FunctionOutputAPI
from ._api.generator import GeneratorAPI
from ._api.generator_efficiency_curve import GeneratorEfficiencyCurveAPI
from ._api.market_configuration import MarketConfigurationAPI
from ._api.multi_scenario_partial_bid_matrix_calculation_input import MultiScenarioPartialBidMatrixCalculationInputAPI
from ._api.partial_bid_configuration import PartialBidConfigurationAPI
from ._api.partial_bid_matrix_calculation_input import PartialBidMatrixCalculationInputAPI
from ._api.partial_bid_matrix_calculation_output import PartialBidMatrixCalculationOutputAPI
from ._api.partial_bid_matrix_information import PartialBidMatrixInformationAPI
from ._api.partial_bid_matrix_information_with_scenarios import PartialBidMatrixInformationWithScenariosAPI
from ._api.plant import PlantAPI
from ._api.plant_information import PlantInformationAPI
from ._api.plant_water_value_based import PlantWaterValueBasedAPI
from ._api.power_asset import PowerAssetAPI
from ._api.price_area import PriceAreaAPI
from ._api.price_area_afrr import PriceAreaAFRRAPI
from ._api.price_area_day_ahead import PriceAreaDayAheadAPI
from ._api.price_area_information import PriceAreaInformationAPI
from ._api.price_production import PriceProductionAPI
from ._api.shop_attribute_mapping import ShopAttributeMappingAPI
from ._api.shop_based_partial_bid_configuration import ShopBasedPartialBidConfigurationAPI
from ._api.shop_case import ShopCaseAPI
from ._api.shop_commands import ShopCommandsAPI
from ._api.shop_file import ShopFileAPI
from ._api.shop_model import ShopModelAPI
from ._api.shop_output_time_series_definition import ShopOutputTimeSeriesDefinitionAPI
from ._api.shop_penalty_report import ShopPenaltyReportAPI
from ._api.shop_preprocessor_input import ShopPreprocessorInputAPI
from ._api.shop_preprocessor_output import ShopPreprocessorOutputAPI
from ._api.shop_result import ShopResultAPI
from ._api.shop_scenario import ShopScenarioAPI
from ._api.shop_scenario_set import ShopScenarioSetAPI
from ._api.shop_time_series import ShopTimeSeriesAPI
from ._api.shop_trigger_input import ShopTriggerInputAPI
from ._api.shop_trigger_output import ShopTriggerOutputAPI
from ._api.task_dispatcher_input import TaskDispatcherInputAPI
from ._api.task_dispatcher_output import TaskDispatcherOutputAPI
from ._api.total_bid_matrix_calculation_input import TotalBidMatrixCalculationInputAPI
from ._api.total_bid_matrix_calculation_output import TotalBidMatrixCalculationOutputAPI
from ._api.turbine_efficiency_curve import TurbineEfficiencyCurveAPI
from ._api.water_value_based_partial_bid_configuration import WaterValueBasedPartialBidConfigurationAPI
from ._api.water_value_based_partial_bid_matrix_calculation_input import (
    WaterValueBasedPartialBidMatrixCalculationInputAPI,
)
from ._api.watercourse import WatercourseAPI
from ._api._core import SequenceNotStr, GraphQLQueryResponse
from .data_classes._core import DEFAULT_INSTANCE_SPACE, GraphQLList
from . import data_classes


class ShopBasedDayAheadBidProcesAPIs:
    """
    ShopBasedDayAheadBidProcesAPIs

    Data Model:
        space: power_ops_core
        externalId: compute_ShopBasedDayAhead
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("power_ops_core", "Alert", "1"),
            data_classes.BidConfigurationDayAhead: dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1"),
            data_classes.BidMatrix: dm.ViewId("power_ops_core", "BidMatrix", "1"),
            data_classes.DateSpecification: dm.ViewId("power_ops_core", "DateSpecification", "1"),
            data_classes.FunctionInput: dm.ViewId("power_ops_core", "FunctionInput", "1"),
            data_classes.FunctionOutput: dm.ViewId("power_ops_core", "FunctionOutput", "1"),
            data_classes.MarketConfiguration: dm.ViewId("power_ops_core", "MarketConfiguration", "1"),
            data_classes.MultiScenarioPartialBidMatrixCalculationInput: dm.ViewId(
                "power_ops_core", "MultiScenarioPartialBidMatrixCalculationInput", "1"
            ),
            data_classes.PartialBidConfiguration: dm.ViewId("power_ops_core", "PartialBidConfiguration", "1"),
            data_classes.PartialBidMatrixCalculationInput: dm.ViewId(
                "power_ops_core", "PartialBidMatrixCalculationInput", "1"
            ),
            data_classes.PartialBidMatrixCalculationOutput: dm.ViewId(
                "power_ops_core", "PartialBidMatrixCalculationOutput", "1"
            ),
            data_classes.PowerAsset: dm.ViewId("power_ops_core", "PowerAsset", "1"),
            data_classes.PriceArea: dm.ViewId("power_ops_core", "PriceArea", "1"),
            data_classes.PriceAreaDayAhead: dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1"),
            data_classes.PriceProduction: dm.ViewId("power_ops_core", "PriceProduction", "1"),
            data_classes.ShopAttributeMapping: dm.ViewId("power_ops_core", "ShopAttributeMapping", "1"),
            data_classes.ShopBasedPartialBidConfiguration: dm.ViewId(
                "power_ops_core", "ShopBasedPartialBidConfiguration", "1"
            ),
            data_classes.ShopCase: dm.ViewId("power_ops_core", "ShopCase", "1"),
            data_classes.ShopCommands: dm.ViewId("power_ops_core", "ShopCommands", "1"),
            data_classes.ShopFile: dm.ViewId("power_ops_core", "ShopFile", "1"),
            data_classes.ShopModel: dm.ViewId("power_ops_core", "ShopModel", "1"),
            data_classes.ShopOutputTimeSeriesDefinition: dm.ViewId(
                "power_ops_core", "ShopOutputTimeSeriesDefinition", "1"
            ),
            data_classes.ShopPreprocessorInput: dm.ViewId("power_ops_core", "ShopPreprocessorInput", "1"),
            data_classes.ShopPreprocessorOutput: dm.ViewId("power_ops_core", "ShopPreprocessorOutput", "1"),
            data_classes.ShopResult: dm.ViewId("power_ops_core", "ShopResult", "1"),
            data_classes.ShopScenario: dm.ViewId("power_ops_core", "ShopScenario", "1"),
            data_classes.ShopScenarioSet: dm.ViewId("power_ops_core", "ShopScenarioSet", "1"),
            data_classes.ShopTimeSeries: dm.ViewId("power_ops_core", "ShopTimeSeries", "1"),
            data_classes.ShopTriggerInput: dm.ViewId("power_ops_core", "ShopTriggerInput", "1"),
            data_classes.ShopTriggerOutput: dm.ViewId("power_ops_core", "ShopTriggerOutput", "1"),
            data_classes.TaskDispatcherInput: dm.ViewId("power_ops_core", "TaskDispatcherInput", "1"),
            data_classes.TaskDispatcherOutput: dm.ViewId("power_ops_core", "TaskDispatcherOutput", "1"),
        }
        self._view_by_read_class = view_by_read_class
        self._client = client

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_configuration_day_ahead = BidConfigurationDayAheadAPI(client, view_by_read_class)
        self.bid_matrix = BidMatrixAPI(client, view_by_read_class)
        self.date_specification = DateSpecificationAPI(client, view_by_read_class)
        self.function_input = FunctionInputAPI(client, view_by_read_class)
        self.function_output = FunctionOutputAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.multi_scenario_partial_bid_matrix_calculation_input = MultiScenarioPartialBidMatrixCalculationInputAPI(
            client, view_by_read_class
        )
        self.partial_bid_configuration = PartialBidConfigurationAPI(client, view_by_read_class)
        self.partial_bid_matrix_calculation_input = PartialBidMatrixCalculationInputAPI(client, view_by_read_class)
        self.partial_bid_matrix_calculation_output = PartialBidMatrixCalculationOutputAPI(client, view_by_read_class)
        self.power_asset = PowerAssetAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_area_day_ahead = PriceAreaDayAheadAPI(client, view_by_read_class)
        self.price_production = PriceProductionAPI(client, view_by_read_class)
        self.shop_attribute_mapping = ShopAttributeMappingAPI(client, view_by_read_class)
        self.shop_based_partial_bid_configuration = ShopBasedPartialBidConfigurationAPI(client, view_by_read_class)
        self.shop_case = ShopCaseAPI(client, view_by_read_class)
        self.shop_commands = ShopCommandsAPI(client, view_by_read_class)
        self.shop_file = ShopFileAPI(client, view_by_read_class)
        self.shop_model = ShopModelAPI(client, view_by_read_class)
        self.shop_output_time_series_definition = ShopOutputTimeSeriesDefinitionAPI(client, view_by_read_class)
        self.shop_preprocessor_input = ShopPreprocessorInputAPI(client, view_by_read_class)
        self.shop_preprocessor_output = ShopPreprocessorOutputAPI(client, view_by_read_class)
        self.shop_result = ShopResultAPI(client, view_by_read_class)
        self.shop_scenario = ShopScenarioAPI(client, view_by_read_class)
        self.shop_scenario_set = ShopScenarioSetAPI(client, view_by_read_class)
        self.shop_time_series = ShopTimeSeriesAPI(client, view_by_read_class)
        self.shop_trigger_input = ShopTriggerInputAPI(client, view_by_read_class)
        self.shop_trigger_output = ShopTriggerOutputAPI(client, view_by_read_class)
        self.task_dispatcher_input = TaskDispatcherInputAPI(client, view_by_read_class)
        self.task_dispatcher_output = TaskDispatcherOutputAPI(client, view_by_read_class)

    def graphql_query(self, query: str, variables: dict[str, Any] | None = None) -> GraphQLList:
        """Execute a GraphQl query against the compute_ShopBasedDayAhead data model.

        Args:
            query (str): The GraphQL query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.
        """
        data_model_id = dm.DataModelId("power_ops_core", "compute_ShopBasedDayAhead", "1")
        result = self._client.data_modeling.graphql.query(data_model_id, query, variables)
        return GraphQLQueryResponse(data_model_id).parse(result)


class TotalBidMatrixCalculationAPIs:
    """
    TotalBidMatrixCalculationAPIs

    Data Model:
        space: power_ops_core
        externalId: compute_TotalBidMatrixCalculation
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("power_ops_core", "Alert", "1"),
            data_classes.BidConfigurationDayAhead: dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1"),
            data_classes.BidDocument: dm.ViewId("power_ops_core", "BidDocument", "1"),
            data_classes.BidDocumentDayAhead: dm.ViewId("power_ops_core", "BidDocumentDayAhead", "1"),
            data_classes.BidMatrix: dm.ViewId("power_ops_core", "BidMatrix", "1"),
            data_classes.BidMatrixInformation: dm.ViewId("power_ops_core", "BidMatrixInformation", "1"),
            data_classes.DateSpecification: dm.ViewId("power_ops_core", "DateSpecification", "1"),
            data_classes.FunctionInput: dm.ViewId("power_ops_core", "FunctionInput", "1"),
            data_classes.FunctionOutput: dm.ViewId("power_ops_core", "FunctionOutput", "1"),
            data_classes.MarketConfiguration: dm.ViewId("power_ops_core", "MarketConfiguration", "1"),
            data_classes.PartialBidConfiguration: dm.ViewId("power_ops_core", "PartialBidConfiguration", "1"),
            data_classes.PartialBidMatrixInformation: dm.ViewId("power_ops_core", "PartialBidMatrixInformation", "1"),
            data_classes.PowerAsset: dm.ViewId("power_ops_core", "PowerAsset", "1"),
            data_classes.PriceArea: dm.ViewId("power_ops_core", "PriceArea", "1"),
            data_classes.PriceAreaDayAhead: dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1"),
            data_classes.PriceProduction: dm.ViewId("power_ops_core", "PriceProduction", "1"),
            data_classes.ShopAttributeMapping: dm.ViewId("power_ops_core", "ShopAttributeMapping", "1"),
            data_classes.ShopCase: dm.ViewId("power_ops_core", "ShopCase", "1"),
            data_classes.ShopCommands: dm.ViewId("power_ops_core", "ShopCommands", "1"),
            data_classes.ShopFile: dm.ViewId("power_ops_core", "ShopFile", "1"),
            data_classes.ShopModel: dm.ViewId("power_ops_core", "ShopModel", "1"),
            data_classes.ShopOutputTimeSeriesDefinition: dm.ViewId(
                "power_ops_core", "ShopOutputTimeSeriesDefinition", "1"
            ),
            data_classes.ShopResult: dm.ViewId("power_ops_core", "ShopResult", "1"),
            data_classes.ShopScenario: dm.ViewId("power_ops_core", "ShopScenario", "1"),
            data_classes.ShopTimeSeries: dm.ViewId("power_ops_core", "ShopTimeSeries", "1"),
            data_classes.TotalBidMatrixCalculationInput: dm.ViewId(
                "power_ops_core", "TotalBidMatrixCalculationInput", "1"
            ),
            data_classes.TotalBidMatrixCalculationOutput: dm.ViewId(
                "power_ops_core", "TotalBidMatrixCalculationOutput", "1"
            ),
        }
        self._view_by_read_class = view_by_read_class
        self._client = client

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_configuration_day_ahead = BidConfigurationDayAheadAPI(client, view_by_read_class)
        self.bid_document = BidDocumentAPI(client, view_by_read_class)
        self.bid_document_day_ahead = BidDocumentDayAheadAPI(client, view_by_read_class)
        self.bid_matrix = BidMatrixAPI(client, view_by_read_class)
        self.bid_matrix_information = BidMatrixInformationAPI(client, view_by_read_class)
        self.date_specification = DateSpecificationAPI(client, view_by_read_class)
        self.function_input = FunctionInputAPI(client, view_by_read_class)
        self.function_output = FunctionOutputAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.partial_bid_configuration = PartialBidConfigurationAPI(client, view_by_read_class)
        self.partial_bid_matrix_information = PartialBidMatrixInformationAPI(client, view_by_read_class)
        self.power_asset = PowerAssetAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_area_day_ahead = PriceAreaDayAheadAPI(client, view_by_read_class)
        self.price_production = PriceProductionAPI(client, view_by_read_class)
        self.shop_attribute_mapping = ShopAttributeMappingAPI(client, view_by_read_class)
        self.shop_case = ShopCaseAPI(client, view_by_read_class)
        self.shop_commands = ShopCommandsAPI(client, view_by_read_class)
        self.shop_file = ShopFileAPI(client, view_by_read_class)
        self.shop_model = ShopModelAPI(client, view_by_read_class)
        self.shop_output_time_series_definition = ShopOutputTimeSeriesDefinitionAPI(client, view_by_read_class)
        self.shop_result = ShopResultAPI(client, view_by_read_class)
        self.shop_scenario = ShopScenarioAPI(client, view_by_read_class)
        self.shop_time_series = ShopTimeSeriesAPI(client, view_by_read_class)
        self.total_bid_matrix_calculation_input = TotalBidMatrixCalculationInputAPI(client, view_by_read_class)
        self.total_bid_matrix_calculation_output = TotalBidMatrixCalculationOutputAPI(client, view_by_read_class)

    def graphql_query(self, query: str, variables: dict[str, Any] | None = None) -> GraphQLList:
        """Execute a GraphQl query against the compute_TotalBidMatrixCalculation data model.

        Args:
            query (str): The GraphQL query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.
        """
        data_model_id = dm.DataModelId("power_ops_core", "compute_TotalBidMatrixCalculation", "1")
        result = self._client.data_modeling.graphql.query(data_model_id, query, variables)
        return GraphQLQueryResponse(data_model_id).parse(result)


class WaterValueBasedDayAheadBidProcesAPIs:
    """
    WaterValueBasedDayAheadBidProcesAPIs

    Data Model:
        space: power_ops_core
        externalId: compute_WaterValueBasedDayAheadBid
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("power_ops_core", "Alert", "1"),
            data_classes.BidConfigurationDayAhead: dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1"),
            data_classes.BidMatrix: dm.ViewId("power_ops_core", "BidMatrix", "1"),
            data_classes.DateSpecification: dm.ViewId("power_ops_core", "DateSpecification", "1"),
            data_classes.FunctionInput: dm.ViewId("power_ops_core", "FunctionInput", "1"),
            data_classes.FunctionOutput: dm.ViewId("power_ops_core", "FunctionOutput", "1"),
            data_classes.Generator: dm.ViewId("power_ops_core", "Generator", "1"),
            data_classes.GeneratorEfficiencyCurve: dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1"),
            data_classes.MarketConfiguration: dm.ViewId("power_ops_core", "MarketConfiguration", "1"),
            data_classes.PartialBidConfiguration: dm.ViewId("power_ops_core", "PartialBidConfiguration", "1"),
            data_classes.PartialBidMatrixCalculationInput: dm.ViewId(
                "power_ops_core", "PartialBidMatrixCalculationInput", "1"
            ),
            data_classes.PartialBidMatrixCalculationOutput: dm.ViewId(
                "power_ops_core", "PartialBidMatrixCalculationOutput", "1"
            ),
            data_classes.Plant: dm.ViewId("power_ops_core", "Plant", "1"),
            data_classes.PlantWaterValueBased: dm.ViewId("power_ops_core", "PlantWaterValueBased", "1"),
            data_classes.PowerAsset: dm.ViewId("power_ops_core", "PowerAsset", "1"),
            data_classes.PriceArea: dm.ViewId("power_ops_core", "PriceArea", "1"),
            data_classes.PriceAreaDayAhead: dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1"),
            data_classes.TaskDispatcherInput: dm.ViewId("power_ops_core", "TaskDispatcherInput", "1"),
            data_classes.TaskDispatcherOutput: dm.ViewId("power_ops_core", "TaskDispatcherOutput", "1"),
            data_classes.TurbineEfficiencyCurve: dm.ViewId("power_ops_core", "TurbineEfficiencyCurve", "1"),
            data_classes.WaterValueBasedPartialBidConfiguration: dm.ViewId(
                "power_ops_core", "WaterValueBasedPartialBidConfiguration", "1"
            ),
            data_classes.WaterValueBasedPartialBidMatrixCalculationInput: dm.ViewId(
                "power_ops_core", "WaterValueBasedPartialBidMatrixCalculationInput", "1"
            ),
        }
        self._view_by_read_class = view_by_read_class
        self._client = client

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_configuration_day_ahead = BidConfigurationDayAheadAPI(client, view_by_read_class)
        self.bid_matrix = BidMatrixAPI(client, view_by_read_class)
        self.date_specification = DateSpecificationAPI(client, view_by_read_class)
        self.function_input = FunctionInputAPI(client, view_by_read_class)
        self.function_output = FunctionOutputAPI(client, view_by_read_class)
        self.generator = GeneratorAPI(client, view_by_read_class)
        self.generator_efficiency_curve = GeneratorEfficiencyCurveAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.partial_bid_configuration = PartialBidConfigurationAPI(client, view_by_read_class)
        self.partial_bid_matrix_calculation_input = PartialBidMatrixCalculationInputAPI(client, view_by_read_class)
        self.partial_bid_matrix_calculation_output = PartialBidMatrixCalculationOutputAPI(client, view_by_read_class)
        self.plant = PlantAPI(client, view_by_read_class)
        self.plant_water_value_based = PlantWaterValueBasedAPI(client, view_by_read_class)
        self.power_asset = PowerAssetAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_area_day_ahead = PriceAreaDayAheadAPI(client, view_by_read_class)
        self.task_dispatcher_input = TaskDispatcherInputAPI(client, view_by_read_class)
        self.task_dispatcher_output = TaskDispatcherOutputAPI(client, view_by_read_class)
        self.turbine_efficiency_curve = TurbineEfficiencyCurveAPI(client, view_by_read_class)
        self.water_value_based_partial_bid_configuration = WaterValueBasedPartialBidConfigurationAPI(
            client, view_by_read_class
        )
        self.water_value_based_partial_bid_matrix_calculation_input = (
            WaterValueBasedPartialBidMatrixCalculationInputAPI(client, view_by_read_class)
        )

    def graphql_query(self, query: str, variables: dict[str, Any] | None = None) -> GraphQLList:
        """Execute a GraphQl query against the compute_WaterValueBasedDayAheadBid data model.

        Args:
            query (str): The GraphQL query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.
        """
        data_model_id = dm.DataModelId("power_ops_core", "compute_WaterValueBasedDayAheadBid", "1")
        result = self._client.data_modeling.graphql.query(data_model_id, query, variables)
        return GraphQLQueryResponse(data_model_id).parse(result)


class DayAheadConfigurationAPIs:
    """
    DayAheadConfigurationAPIs

    Data Model:
        space: power_ops_core
        externalId: config_DayAheadConfiguration
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.BidConfigurationDayAhead: dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1"),
            data_classes.DateSpecification: dm.ViewId("power_ops_core", "DateSpecification", "1"),
            data_classes.Generator: dm.ViewId("power_ops_core", "Generator", "1"),
            data_classes.GeneratorEfficiencyCurve: dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1"),
            data_classes.MarketConfiguration: dm.ViewId("power_ops_core", "MarketConfiguration", "1"),
            data_classes.PartialBidConfiguration: dm.ViewId("power_ops_core", "PartialBidConfiguration", "1"),
            data_classes.Plant: dm.ViewId("power_ops_core", "Plant", "1"),
            data_classes.PlantInformation: dm.ViewId("power_ops_core", "PlantInformation", "1"),
            data_classes.PlantWaterValueBased: dm.ViewId("power_ops_core", "PlantWaterValueBased", "1"),
            data_classes.PowerAsset: dm.ViewId("power_ops_core", "PowerAsset", "1"),
            data_classes.PriceArea: dm.ViewId("power_ops_core", "PriceArea", "1"),
            data_classes.PriceAreaDayAhead: dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1"),
            data_classes.ShopAttributeMapping: dm.ViewId("power_ops_core", "ShopAttributeMapping", "1"),
            data_classes.ShopBasedPartialBidConfiguration: dm.ViewId(
                "power_ops_core", "ShopBasedPartialBidConfiguration", "1"
            ),
            data_classes.ShopCommands: dm.ViewId("power_ops_core", "ShopCommands", "1"),
            data_classes.ShopFile: dm.ViewId("power_ops_core", "ShopFile", "1"),
            data_classes.ShopModel: dm.ViewId("power_ops_core", "ShopModel", "1"),
            data_classes.ShopOutputTimeSeriesDefinition: dm.ViewId(
                "power_ops_core", "ShopOutputTimeSeriesDefinition", "1"
            ),
            data_classes.ShopScenario: dm.ViewId("power_ops_core", "ShopScenario", "1"),
            data_classes.ShopScenarioSet: dm.ViewId("power_ops_core", "ShopScenarioSet", "1"),
            data_classes.TurbineEfficiencyCurve: dm.ViewId("power_ops_core", "TurbineEfficiencyCurve", "1"),
            data_classes.WaterValueBasedPartialBidConfiguration: dm.ViewId(
                "power_ops_core", "WaterValueBasedPartialBidConfiguration", "1"
            ),
        }
        self._view_by_read_class = view_by_read_class
        self._client = client

        self.bid_configuration_day_ahead = BidConfigurationDayAheadAPI(client, view_by_read_class)
        self.date_specification = DateSpecificationAPI(client, view_by_read_class)
        self.generator = GeneratorAPI(client, view_by_read_class)
        self.generator_efficiency_curve = GeneratorEfficiencyCurveAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.partial_bid_configuration = PartialBidConfigurationAPI(client, view_by_read_class)
        self.plant = PlantAPI(client, view_by_read_class)
        self.plant_information = PlantInformationAPI(client, view_by_read_class)
        self.plant_water_value_based = PlantWaterValueBasedAPI(client, view_by_read_class)
        self.power_asset = PowerAssetAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_area_day_ahead = PriceAreaDayAheadAPI(client, view_by_read_class)
        self.shop_attribute_mapping = ShopAttributeMappingAPI(client, view_by_read_class)
        self.shop_based_partial_bid_configuration = ShopBasedPartialBidConfigurationAPI(client, view_by_read_class)
        self.shop_commands = ShopCommandsAPI(client, view_by_read_class)
        self.shop_file = ShopFileAPI(client, view_by_read_class)
        self.shop_model = ShopModelAPI(client, view_by_read_class)
        self.shop_output_time_series_definition = ShopOutputTimeSeriesDefinitionAPI(client, view_by_read_class)
        self.shop_scenario = ShopScenarioAPI(client, view_by_read_class)
        self.shop_scenario_set = ShopScenarioSetAPI(client, view_by_read_class)
        self.turbine_efficiency_curve = TurbineEfficiencyCurveAPI(client, view_by_read_class)
        self.water_value_based_partial_bid_configuration = WaterValueBasedPartialBidConfigurationAPI(
            client, view_by_read_class
        )

    def graphql_query(self, query: str, variables: dict[str, Any] | None = None) -> GraphQLList:
        """Execute a GraphQl query against the config_DayAheadConfiguration data model.

        Args:
            query (str): The GraphQL query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.
        """
        data_model_id = dm.DataModelId("power_ops_core", "config_DayAheadConfiguration", "1")
        result = self._client.data_modeling.graphql.query(data_model_id, query, variables)
        return GraphQLQueryResponse(data_model_id).parse(result)


class AFRRBidAPIs:
    """
    AFRRBidAPIs

    Data Model:
        space: power_ops_core
        externalId: frontend_AFRRBid
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("power_ops_core", "Alert", "1"),
            data_classes.BidDocument: dm.ViewId("power_ops_core", "BidDocument", "1"),
            data_classes.BidDocumentAFRR: dm.ViewId("power_ops_core", "BidDocumentAFRR", "1"),
            data_classes.BidRow: dm.ViewId("power_ops_core", "BidRow", "1"),
            data_classes.PowerAsset: dm.ViewId("power_ops_core", "PowerAsset", "1"),
            data_classes.PriceArea: dm.ViewId("power_ops_core", "PriceArea", "1"),
            data_classes.PriceAreaAFRR: dm.ViewId("power_ops_core", "PriceAreaAFRR", "1"),
        }
        self._view_by_read_class = view_by_read_class
        self._client = client

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_document = BidDocumentAPI(client, view_by_read_class)
        self.bid_document_afrr = BidDocumentAFRRAPI(client, view_by_read_class)
        self.bid_row = BidRowAPI(client, view_by_read_class)
        self.power_asset = PowerAssetAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_area_afrr = PriceAreaAFRRAPI(client, view_by_read_class)

    def graphql_query(self, query: str, variables: dict[str, Any] | None = None) -> GraphQLList:
        """Execute a GraphQl query against the frontend_AFRRBid data model.

        Args:
            query (str): The GraphQL query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.
        """
        data_model_id = dm.DataModelId("power_ops_core", "frontend_AFRRBid", "1")
        result = self._client.data_modeling.graphql.query(data_model_id, query, variables)
        return GraphQLQueryResponse(data_model_id).parse(result)


class PowerAssetAPIs:
    """
    PowerAssetAPIs

    Data Model:
        space: power_ops_core
        externalId: frontend_Asset
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.BidConfigurationDayAhead: dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1"),
            data_classes.DateSpecification: dm.ViewId("power_ops_core", "DateSpecification", "1"),
            data_classes.Generator: dm.ViewId("power_ops_core", "Generator", "1"),
            data_classes.GeneratorEfficiencyCurve: dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1"),
            data_classes.MarketConfiguration: dm.ViewId("power_ops_core", "MarketConfiguration", "1"),
            data_classes.PartialBidConfiguration: dm.ViewId("power_ops_core", "PartialBidConfiguration", "1"),
            data_classes.Plant: dm.ViewId("power_ops_core", "Plant", "1"),
            data_classes.PlantInformation: dm.ViewId("power_ops_core", "PlantInformation", "1"),
            data_classes.PlantWaterValueBased: dm.ViewId("power_ops_core", "PlantWaterValueBased", "1"),
            data_classes.PowerAsset: dm.ViewId("power_ops_core", "PowerAsset", "1"),
            data_classes.PriceArea: dm.ViewId("power_ops_core", "PriceArea", "1"),
            data_classes.PriceAreaAFRR: dm.ViewId("power_ops_core", "PriceAreaAFRR", "1"),
            data_classes.PriceAreaDayAhead: dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1"),
            data_classes.PriceAreaInformation: dm.ViewId("power_ops_core", "PriceAreaInformation", "1"),
            data_classes.TurbineEfficiencyCurve: dm.ViewId("power_ops_core", "TurbineEfficiencyCurve", "1"),
            data_classes.Watercourse: dm.ViewId("power_ops_core", "Watercourse", "1"),
        }
        self._view_by_read_class = view_by_read_class
        self._client = client

        self.bid_configuration_day_ahead = BidConfigurationDayAheadAPI(client, view_by_read_class)
        self.date_specification = DateSpecificationAPI(client, view_by_read_class)
        self.generator = GeneratorAPI(client, view_by_read_class)
        self.generator_efficiency_curve = GeneratorEfficiencyCurveAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.partial_bid_configuration = PartialBidConfigurationAPI(client, view_by_read_class)
        self.plant = PlantAPI(client, view_by_read_class)
        self.plant_information = PlantInformationAPI(client, view_by_read_class)
        self.plant_water_value_based = PlantWaterValueBasedAPI(client, view_by_read_class)
        self.power_asset = PowerAssetAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_area_afrr = PriceAreaAFRRAPI(client, view_by_read_class)
        self.price_area_day_ahead = PriceAreaDayAheadAPI(client, view_by_read_class)
        self.price_area_information = PriceAreaInformationAPI(client, view_by_read_class)
        self.turbine_efficiency_curve = TurbineEfficiencyCurveAPI(client, view_by_read_class)
        self.watercourse = WatercourseAPI(client, view_by_read_class)

    def graphql_query(self, query: str, variables: dict[str, Any] | None = None) -> GraphQLList:
        """Execute a GraphQl query against the frontend_Asset data model.

        Args:
            query (str): The GraphQL query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.
        """
        data_model_id = dm.DataModelId("power_ops_core", "frontend_Asset", "1")
        result = self._client.data_modeling.graphql.query(data_model_id, query, variables)
        return GraphQLQueryResponse(data_model_id).parse(result)


class DayAheadBidAPIs:
    """
    DayAheadBidAPIs

    Data Model:
        space: power_ops_core
        externalId: frontend_DayAheadBid
        version: 1

    """

    def __init__(self, client: CogniteClient):
        view_by_read_class = {
            data_classes.Alert: dm.ViewId("power_ops_core", "Alert", "1"),
            data_classes.BidConfigurationDayAhead: dm.ViewId("power_ops_core", "BidConfigurationDayAhead", "1"),
            data_classes.BidDocument: dm.ViewId("power_ops_core", "BidDocument", "1"),
            data_classes.BidDocumentDayAhead: dm.ViewId("power_ops_core", "BidDocumentDayAhead", "1"),
            data_classes.BidMatrix: dm.ViewId("power_ops_core", "BidMatrix", "1"),
            data_classes.BidMatrixInformation: dm.ViewId("power_ops_core", "BidMatrixInformation", "1"),
            data_classes.DateSpecification: dm.ViewId("power_ops_core", "DateSpecification", "1"),
            data_classes.MarketConfiguration: dm.ViewId("power_ops_core", "MarketConfiguration", "1"),
            data_classes.PartialBidConfiguration: dm.ViewId("power_ops_core", "PartialBidConfiguration", "1"),
            data_classes.PartialBidMatrixInformation: dm.ViewId("power_ops_core", "PartialBidMatrixInformation", "1"),
            data_classes.PartialBidMatrixInformationWithScenarios: dm.ViewId(
                "power_ops_core", "PartialBidMatrixInformationWithScenarios", "1"
            ),
            data_classes.PowerAsset: dm.ViewId("power_ops_core", "PowerAsset", "1"),
            data_classes.PriceArea: dm.ViewId("power_ops_core", "PriceArea", "1"),
            data_classes.PriceAreaDayAhead: dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1"),
            data_classes.PriceProduction: dm.ViewId("power_ops_core", "PriceProduction", "1"),
            data_classes.ShopAttributeMapping: dm.ViewId("power_ops_core", "ShopAttributeMapping", "1"),
            data_classes.ShopCase: dm.ViewId("power_ops_core", "ShopCase", "1"),
            data_classes.ShopCommands: dm.ViewId("power_ops_core", "ShopCommands", "1"),
            data_classes.ShopFile: dm.ViewId("power_ops_core", "ShopFile", "1"),
            data_classes.ShopModel: dm.ViewId("power_ops_core", "ShopModel", "1"),
            data_classes.ShopOutputTimeSeriesDefinition: dm.ViewId(
                "power_ops_core", "ShopOutputTimeSeriesDefinition", "1"
            ),
            data_classes.ShopPenaltyReport: dm.ViewId("power_ops_core", "ShopPenaltyReport", "1"),
            data_classes.ShopResult: dm.ViewId("power_ops_core", "ShopResult", "1"),
            data_classes.ShopScenario: dm.ViewId("power_ops_core", "ShopScenario", "1"),
            data_classes.ShopTimeSeries: dm.ViewId("power_ops_core", "ShopTimeSeries", "1"),
        }
        self._view_by_read_class = view_by_read_class
        self._client = client

        self.alert = AlertAPI(client, view_by_read_class)
        self.bid_configuration_day_ahead = BidConfigurationDayAheadAPI(client, view_by_read_class)
        self.bid_document = BidDocumentAPI(client, view_by_read_class)
        self.bid_document_day_ahead = BidDocumentDayAheadAPI(client, view_by_read_class)
        self.bid_matrix = BidMatrixAPI(client, view_by_read_class)
        self.bid_matrix_information = BidMatrixInformationAPI(client, view_by_read_class)
        self.date_specification = DateSpecificationAPI(client, view_by_read_class)
        self.market_configuration = MarketConfigurationAPI(client, view_by_read_class)
        self.partial_bid_configuration = PartialBidConfigurationAPI(client, view_by_read_class)
        self.partial_bid_matrix_information = PartialBidMatrixInformationAPI(client, view_by_read_class)
        self.partial_bid_matrix_information_with_scenarios = PartialBidMatrixInformationWithScenariosAPI(
            client, view_by_read_class
        )
        self.power_asset = PowerAssetAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.price_area_day_ahead = PriceAreaDayAheadAPI(client, view_by_read_class)
        self.price_production = PriceProductionAPI(client, view_by_read_class)
        self.shop_attribute_mapping = ShopAttributeMappingAPI(client, view_by_read_class)
        self.shop_case = ShopCaseAPI(client, view_by_read_class)
        self.shop_commands = ShopCommandsAPI(client, view_by_read_class)
        self.shop_file = ShopFileAPI(client, view_by_read_class)
        self.shop_model = ShopModelAPI(client, view_by_read_class)
        self.shop_output_time_series_definition = ShopOutputTimeSeriesDefinitionAPI(client, view_by_read_class)
        self.shop_penalty_report = ShopPenaltyReportAPI(client, view_by_read_class)
        self.shop_result = ShopResultAPI(client, view_by_read_class)
        self.shop_scenario = ShopScenarioAPI(client, view_by_read_class)
        self.shop_time_series = ShopTimeSeriesAPI(client, view_by_read_class)

    def graphql_query(self, query: str, variables: dict[str, Any] | None = None) -> GraphQLList:
        """Execute a GraphQl query against the frontend_DayAheadBid data model.

        Args:
            query (str): The GraphQL query to issue.
            variables (dict[str, Any] | None): An optional dict of variables to pass to the query.
        """
        data_model_id = dm.DataModelId("power_ops_core", "frontend_DayAheadBid", "1")
        result = self._client.data_modeling.graphql.query(data_model_id, query, variables)
        return GraphQLQueryResponse(data_model_id).parse(result)


class PowerOpsModelsV1Client:
    """
    PowerOpsModelsV1Client

    Generated with:
        pygen = 0.99.24
        cognite-sdk = 7.46.1
        pydantic = 2.7.2

    """

    def __init__(self, config_or_client: CogniteClient | ClientConfig):
        if isinstance(config_or_client, CogniteClient):
            client = config_or_client
        elif isinstance(config_or_client, ClientConfig):
            client = CogniteClient(config_or_client)
        else:
            raise ValueError(f"Expected CogniteClient or ClientConfig, got {type(config_or_client)}")
        # The client name is used for aggregated logging of Pygen Usage
        client.config.client_name = "CognitePygen:0.99.24"

        self.shop_based_day_ahead_bid_process = ShopBasedDayAheadBidProcesAPIs(client)
        self.total_bid_matrix_calculation = TotalBidMatrixCalculationAPIs(client)
        self.water_value_based_day_ahead_bid_process = WaterValueBasedDayAheadBidProcesAPIs(client)
        self.day_ahead_configuration = DayAheadConfigurationAPIs(client)
        self.afrr_bid = AFRRBidAPIs(client)
        self.power_asset = PowerAssetAPIs(client)
        self.day_ahead_bid = DayAheadBidAPIs(client)

        self._client = client
        self._view_by_read_class = {
            k: v
            for api in [
                self.shop_based_day_ahead_bid_process,
                self.total_bid_matrix_calculation,
                self.water_value_based_day_ahead_bid_process,
                self.day_ahead_configuration,
                self.afrr_bid,
                self.power_asset,
                self.day_ahead_bid,
            ]
            for k, v in api._view_by_read_class.items()
        }

    def upsert(
        self,
        items: data_classes.DomainModelWrite | Sequence[data_classes.DomainModelWrite],
        replace: bool = False,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> data_classes.ResourcesWriteResult:
        """Add or update (upsert) items.

        Args:
            items: One or more instances of the pygen generated data classes.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method will, by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
            allow_version_increase (bool): If set to true, the version of the instance will be increased if the instance already exists.
                If you get an error: 'A version conflict caused the ingest to fail', you can set this to true to allow
                the version to increase.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        """
        instances = self._create_instances(items, write_none, allow_version_increase)
        result = self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )
        time_series = []
        if instances.time_series:
            time_series = self._client.time_series.upsert(instances.time_series, mode="patch")

        return data_classes.ResourcesWriteResult(result.nodes, result.edges, TimeSeriesList(time_series))

    def _create_instances(
        self,
        items: data_classes.DomainModelWrite | Sequence[data_classes.DomainModelWrite],
        write_none: bool,
        allow_version_increase: bool,
    ) -> data_classes.ResourcesWrite:
        if isinstance(items, data_classes.DomainModelWrite):
            instances = items.to_instances_write(self._view_by_read_class, write_none, allow_version_increase)
        else:
            instances = data_classes.ResourcesWrite()
            cache: set[tuple[str, str]] = set()
            for item in items:
                instances.extend(
                    item._to_instances_write(
                        cache,
                        self._view_by_read_class,
                        write_none,
                        allow_version_increase,
                    )
                )
        return instances

    def apply(
        self,
        items: data_classes.DomainModelWrite | Sequence[data_classes.DomainModelWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> data_classes.ResourcesWriteResult:
        """Add or update (upsert) items.

        Args:
            items: One or more instances of the pygen generated data classes.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method will, by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the instead."
            "The motivation is that .upsert is a more descriptive name for the operation.",
            UserWarning,
            stacklevel=2,
        )
        return self.upsert(items, replace, write_none)

    def delete(
        self,
        external_id: (
            str | SequenceNotStr[str] | data_classes.DomainModelWrite | Sequence[data_classes.DomainModelWrite]
        ),
        space: str = DEFAULT_INSTANCE_SPACE,
    ) -> dm.InstancesDeleteResult:
        """Delete one or more items.

        If you pass in an item, it will be deleted recursively, i.e., all connected nodes and edges
        will be deleted as well.

        Args:
            external_id: The external id or items(s) to delete.
            space: The space where all the item(s) are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete item by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.delete("my_node_external_id")
        """
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        elif all(isinstance(item, str) for item in external_id):
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )
        elif isinstance(external_id, data_classes.DomainModelWrite) or all(
            isinstance(item, data_classes.DomainModelWrite) for item in external_id
        ):
            resources = self._create_instances(external_id, False, False)
            return self._client.data_modeling.instances.delete(
                nodes=resources.nodes.as_ids(),
                edges=resources.edges.as_ids(),
            )
        else:
            raise ValueError(
                "Expected str, list of str, or DomainModelWrite, list of DomainModelWrite," f"got {type(external_id)}"
            )

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> PowerOpsModelsV1Client:
        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)
        config = ClientConfig.default(project, cdf_cluster, credentials)

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str, section: str | None = "cognite") -> PowerOpsModelsV1Client:
        import toml

        toml_content = toml.load(file_path)
        if section is not None:
            try:
                toml_content = toml_content[section]
            except KeyError as e:
                raise ValueError(f"Could not find section '{section}' in {file_path}") from e

        return cls.azure_project(**toml_content)



import abc
import inspect
from collections.abc import Iterable, Sequence
from typing import TypeVar

T_Type = TypeVar("T_Type", bound=type)


def all_subclasses(base: T_Type) -> list[T_Type]:
    """Returns a list (without duplicates) of all subclasses of a given class, sorted on import-path-name.
    Ignores classes not part of the main library, e.g. subclasses part of tests.
    """
    return sorted(
        filter(
            lambda sub: str(sub).startswith("<class 'cognite.powerops"),
            set(base.__subclasses__()).union(s for c in base.__subclasses__() for s in all_subclasses(c)),
        ),
        key=str,
    )


def all_concrete_subclasses(base: T_Type) -> list[T_Type]:
    return [
        sub
        for sub in all_subclasses(base)
        if all(base is not abc.ABC for base in sub.__bases__) and not inspect.isabstract(sub)
    ]


def chunks(sequence: Sequence[T_Type], chunk_size: int) -> Iterable[Sequence[T_Type]]:
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(sequence), chunk_size):
        yield sequence[i : i + chunk_size]
